;-----------------------------------------------------------------------------
; Z9001-BIOS Version 192K-Modul, 2x800K-LW, 24K-RAMDISK
; Volker Pohlers 11.04.2007
; 07.05.2007 Umstellung auf 16-Bit-Tracks, nderung Adress-Bestimmung 
;-----------------------------------------------------------------------------

		cpu	z80
		page	0

		title	"CPMZ9 BIOS"

FDCPORTS	EQU 	'ROBOTRON'	; Ports 98h, 99h, 0A0h
;FDCPORTS	EQU 	'ROSSENDORF'	; Ports 10h, 11h, 12h

RAMDSK	equ	1			; 0 - keine RAM-Disk

;-----------------------------------------------------------------------------
;
;-----------------------------------------------------------------------------
;                                Bank    Bank
;             CPM-Bank+norm.RAM  RAMDisk CCP-Kopie
;
;     RAMEND ___________________        ______
;     e7ffh |  ramdisk-treiber  |______|______|
;     e000h |                   |      |  
;           |  wboot<ccp-kopie  |      |  
;           |                   |      |  
;           |        BIOS       |      |  
;     UBIOS |___________________| RAM  |
;           |                   | DISK |  
;           |        BDOS       |      |  
;     BDOS  |___________________|      |
;           |                   |      |
;           |        CCP        |      |   You can overwrite this if you want
;     CCP   |___________________|______|
;     c000h |                   |        <-- your stack grows down from here
;           |                   |
;           |                   |
;           |        TPA        |
;           |                   |
;           |   COLD-Start      |
;     4000h |   LOADADRESS      |
;           |                   |
;           |                   |
;     0100h |___________________|        <-- program loaded in here
;           |                   |
;           | SYSTEM PARAMETERS |  
;           |___________________|
;     0000h
;
;Nutzung des 192k-Moduls:
;zur Verfgung stehen RAM C000-DFFF, Bank Cx-Fx und RAM E000-E7FF, Bank x8-xF
;
;C000-DFFF + E000-E7FF, Bank FF	-> komplettes CPM
;C000-DFFF, Bank Cx-Ex	-> RAM-Disk
;E000-E7FF, Bank xC	-> CCP-Kopie


CCP		equ	0C000h		; base of ccp
RAMEND		equ	0E7FFh		; Stackbereich	fr BIOS (nur read disk)
stdiobyte:	equ	11010100b	; I/O-Byte: LIST=UL:,PUNCH=UP1:,READER=UR1:,CONST=TTY:


BDOS       	equ	CCP+0806h	; base of bdos
UBIOS		equ	CCP+1600h	; base of bios
LOADADRESS	equ	4000H		; Hierhin wird das Programm durch den Loader geladen


CTC0		equ	80h	; System CTC0
CTC1		equ	81h	; System CTC1
CTC3		equ	83h	; System CTC3

PIO1AD		equ	88h	; System PIO1AD
PIO1BD		equ	89h	; System PIO1BD	Anwenderport
PIO1BC		equ	8bh	; System PIO1BC Anwenderport

; im 192K-Modul genutzte Banken
bank_base	equ	0eh
bank_cpm	equ	bank_base<<4 | bank_base
bank_ccpkopie	equ	bank_base<<4 | 0dh
bank_rd1	equ	0d0h | bank_base	; ramdisk 1. 8K-Bereich
bank_rd2	equ	0c0h | bank_base	; ramdisk 2. 8K-Bereich
bank_rd3	equ	0f0h | bank_base	; ramdisk 3. 8K-Bereich

; Funktionen fr AS
hi		function x,(x>>8)&255 	; oberes Byte eines 16-Bit-Wortes
nextpage	function x, ((x+0ffh)/100h)*100h

; "Debug"funktion - Schreiben eines Zeichens rechts oben auf den Bildschirm
dbg		macro	char
		push	af
		ld	a,char
		ld	(0ec00h+39),a	; rechts oben
		pop	af
		endm

bank		macro	nr
		ld	a,nr
		out	(0ffh),a
		endm

;-----------------------------------------------------------------------------
; Initialisierung Loader (CPMLDR)
; vom aktuellen Boot-Loader wird dieser Code nicht ausgefhrt, es wird nur
; die Ladeadresse entnommen.
;-----------------------------------------------------------------------------

		org 	100h
anfang:

		di
		ld	de, LOADADRESS	; <- dieser Befehl mu stehen, damit der
					; Loader das CPMZ9 als gltig erkennt
					; Gleichzeitig entnimmt der Loader hier
					; die initiale Ladeadresse.
		ld	hl, ccpstart
		ld	bc, biosend-LOADADRESS
		ldir
;
		xor	a		; Laufwerk 0
		jp	LOADADRESS+biosstart-ccpstart

;-----------------------------------------------------------------------------
; CCP+BDOS
;-----------------------------------------------------------------------------

		align	80h		; 1 Sektor weiter

ccpstart:	equ	$

		phase	CCP


		IF CCP = 09400h
		binclude ccp29400.cpm
		ELSEIF CCP = 09800h
		binclude ccp29800.cpm
		ELSEIF CCP = 0b800h
		binclude ccp2b800.cpm
		ELSEIF CCP = 0c000h
		binclude ccp2c000.cpm
		ELSE
		db	1600h dup (?)	; Bereich CCP+BDOS
		WARNING "CCP \{CCP} nicht erzeugt" 
		ENDIF

		dephase

;-----------------------------------------------------------------------------
; Shadow BIOS
;-----------------------------------------------------------------------------

biosstart:	equ	$

		phase 	UBIOS

;;		jp	cold-CCP+LOADADRESS	; boot im Loadbereich 8000h ff.
		jp	cold		; wird in cold gepatcht durch jp cboot

		jp	wboot
		jp	_const		; const im upper bios
		jp	0F009h		; conin im z9001-os
		jp	_conout		; conout im upper bios
		jp	0F00Fh		; list im z9001-os
		jp	0F012h		; punch im z9001-os
		jp	0F015h		; reader im z9001-os
		jp	home
		jp	seldsk
		jp	settrk
		jp	setsec
		jp	setdma
uread:		jp	_read1
uwrite:		jp	_write1
		jp	0F02Dh		; listst im z9001-os
		jp	sectran

;-----------------------------------------------------------------------------
; PUN-Gertetreiber fr	PUNCH
;-----------------------------------------------------------------------------
upunch:		ld	a, c
		ld	de, upunch1+1
		call	hexa		; A nach hex konvertieren, Eintragen nach (DE),	2x inc DE
		ld	c, '('
		call	_conout		; _conout im upper Bios
upunch1:	ld	bc, 0
		push	bc
		call	_conout		; _conout im upper Bios
		pop	bc
		ld	c, b
		call	_conout		; _conout im upper Bios
		ld	c, ')'
		jp	_conout		; _conout im upper Bios
stopkey:	call	0FD33h		; Z9001-OS DECO0: Abfrage Tastaturmatrix
		ei
		sub	3		; STOP-Taste ?
		or	a
		ret	nz
		ld	(25h), a	; Tastaturbuffer
		scf
		ret

;-----------------------------------------------------------------------------
; TTY-Treiber fr CONS
;-----------------------------------------------------------------------------
uttyc:		cp	1		;
uttyc1:		scf
		ccf
		jp	nz, 0F8F1h	; Z9001-OS CRT:	Steuerprogramm des CRT-Treibers
uttyc2:		call	0F8F1h		; Z9001-OS CRT:	Steuerprogramm des CRT-Treibers
		call	bell
		ld	hl, lstflag

		bit	0, (hl)
		jr	nz, uttyc10
		cp	1Ch		; LIST-Taste
		jr	nz, uttyc10
		ld	a, 1
		ld	(hl), a
		jr	uttyc

uttyc10:	bit	0, (hl)
		jr	z, uttyc4
		ld	(hl), 0
		cp	1Ch		; LIST-Taste
		ret	z
		cp	'P'
		jr	nz, uttyc12
		ld	a, (15h)
		xor	1
		ld	(15h), a
uttyc11:	ld	a, 1
		jr	uttyc

uttyc12:	cp	'N'
		jr	nz, uttyc18
		push	de
		ld	hl, 0EC00h	; Adr. Bildwiederholspeicher
		in	a, (PIO1AD)
		bit	2, a
		ld	a, 14h
		jr	nz, uttyc13
		add	a, 4
uttyc13:	ld	d, a
uttyc14:	ld	e, 28h ; '('
uttyc15:	ld	c, (hl)
		push	de
		push	hl
		call	0F00Fh		; Z9001-OS: list
		pop	hl
		pop	de
		jr	c, uttyc17
		inc	hl
		dec	e
		jr	nz, uttyc15
		push	hl
		push	de
		ld	c, 0Dh
		call	0F00Fh		; Z9001-OS: list
		jr	c, uttyc16
		ld	c, 0Ah
		call	0F00Fh		; Z9001-OS: list
uttyc16:	pop	de
		pop	hl
		jr	c, uttyc17
		dec	d
		jr	nz, uttyc14
uttyc17:	pop	de
		jr	uttyc11
uttyc18:	equ	$

		ld	hl, lsttab	; Tabelle Extrazeichen (LIST+char)
		ld	bc, 7
		cpir			; suche	Sondertaste
		jr	nz, uttyc4
		ld	bc, 6
		add	hl, bc
		ld	a, (hl)		; Sonderzeichen	holen

uttyc4:		ld	hl, uttyc6
		push	hl
		cp	'A'
		jr	c, uttyc5
		cp	'['
		jr	nc, uttyc5
		or	20h
		ret
uttyc5:		cp	'a'
		ret	c
		cp	'{'
		ret	nc
		sub	20h
		ret
uttyc6:	or	a
		ret
lstflag:	db	0

;-----------------------------------------------------------------------------
;05.04.2007 09:56 warum geht das nicht?
;		;align	100h
;		dephase
;		phase	UBIOS+0100h
;;; hier ist ein brauchbarer Platz fr die Interruptvektortabelle
;intvectab:	db	12 dup(?)	; Interruptvektortabelle, 12 Byte
;-----------------------------------------------------------------------------

lsttab:		db	'8'		; Tabelle Extrazeichen (LIST+char)
		db	'9'
		db	','
		db	'.'
		db	'I'
		db	'?'
		db	'='
		db	'['
		db	']'
		db	'{'
		db	'}'
		db	'|'
		db	5Ch		; '\'
		db	'~'

;-----------------------------------------------------------------------------
; CRT-Treiber fr CONS
;-----------------------------------------------------------------------------
ucrtc:		cp	1		;
		jp	nz, uttyc1
		call	0F8F1h		; Z9001-OS CRT:	Steuerprogramm des CRT-Treibers

bell:		di
		push	af
		push	bc
		ld	b, 0
		ld	c, 14h
		ld	a, 00000111b	; interrupt aus, zeitgeber mode, Vorteiler 16, negative	Flange,
					; Start	sofort,	Konstante folgt, Kanal Reset
		out	(CTC0), a	; CTC0
		ld	a, 96h		; Zeitkonstante
		out	(CTC0), a
		in	a, (PIO1AD)
		set	7, a
		out	(PIO1AD), a	; Lautsprecher an
bell1:		djnz	bell1
		dec	c
		jr	nz, bell1
		res	7, a
		out	(PIO1AD), a	; Lautsprecher aus
		ld	a, 00000011b
		out	(CTC0), a	; CTC0 Reset
		pop	bc
		pop	af
		ei
		ret

;-----------------------------------------------------------------------------
; UP2 Gertreiber fr PUNCH
;-----------------------------------------------------------------------------
uup:		push	af
		ld	a, 0Ch
		jr	ucrtl1

;-----------------------------------------------------------------------------
; CRT-Gertetreiber fr	LIST
;-----------------------------------------------------------------------------
ucrtl:		push	af
		ld	a, 7Eh
ucrtl1:		ld	(ucrtl4+1), a
		ld	a, 11001111b
		out	(PIO1BC), a	; PIO1B	init Bitmode
		ld	a, 10000000b
		out	(PIO1BC), a	; Bit7 Eingabe,	Bit6-Bit0 Ausgabe
		pop	af
		inc	a
		ret	z
		dec	a
		jr	nz, ucrtl2
		in	a, (PIO1BD)	; PIO1B	lesen
		or	7Fh
		cpl
		ret
ucrtl2:		call	stopkey
		jr	nc, ucrtl3	; Sprung, wenn STOP-Taste nicht	gedrckt
		ld	(15h), a	; LISW,	Schalter fr Hardcopy
		ret
ucrtl3:		in	a, (PIO1BD)	; PIO1B	lesen
		add	a, a
		jr	c, ucrtl2
		ld	a, c
		cp	7Fh
		jr	nz, ucrtl4
		ld	a, 1Bh
ucrtl4:		ld	e, 7Eh
		ld	b, 9
		di
		or	a
		rla
ucrtl5:		out	(PIO1BD), a	; PIO1B	schreiben
		call	ucrtl6		; kurze	Pause, Zeitwert	in E
		rra
		djnz	ucrtl5
		or	1
		out	(PIO1BD), a	; PIO1B	schreiben
		ei
; kurze	Pause, Zeitwert	in E
ucrtl6:		push	de
ucrtl7:		dec	e
		jr	nz, ucrtl7
		pop	de
		ret

;-----------------------------------------------------------------------------
; UL Gertetreiber fr LIST
;-----------------------------------------------------------------------------
uul:		inc	a
		ret	z
		dec	a
		cpl
		ret	z
		ld	a, 11001111b
		out	(PIO1BC), a
		xor	a
		out	(PIO1BC), a
		ld	a, 01010111b
		out	(CTC1), a
		out	(CTC1), a
		ld	a, c
		or	80h
		cp	0FFh
		jr	nz, uul1
		ld	a, 9Bh
uul1:		out	(PIO1BD), a
		push	ix
		pop	ix
		and	7Fh
		out	(PIO1BD), a
		or	80h
		out	(PIO1BD), a
uul2:		call	stopkey
		ret	c
		in	a, (CTC1)
		cp	57h
		jr	z, uul2
		xor	a
		ret

;-----------------------------------------------------------------------------
;
;-----------------------------------------------------------------------------

;CPA-Source: BIOSDSK.MAC

; Laufwerksauswahl
;=================
; i C: Laufw. (0=A:, 1=B: ...)
; i E: Bit 0 ist LOGIN-Bit von BDOS (=0, wenn LOGIN)

seldsk:		call	dgetpb			; HL auf DPH stellen, IX auf DPB
		ret	z			; Geraet ex. nicht
		ld	a, c
		ld	(ddrive), a		; Laufwerk merken
		bit	0, e			; LOGIN von BDOS?
		ret	nz			; nein

		IF RAMDSK=1
		cp	2		; RAM-DISC C?
		ret	z
		ENDIF

		push	hl			; ^DPH merken
		ld	hl, dbdev
		cp	(hl)			; LOGIN fuer gepuffertes Device?
		jr	nz, seldsk1		; nein
		ld	(hl), 0FFh		; Puffer ist ungueltig
		ld	hl, dbflg
		res	2, (hl)			;kein veraend. Puffer auszug.!

seldsk1:	bit	6, (ix+0Fh)		;Formaterkennung unterdr.?
		jr	z, seldsk1a
		ld	(byte_5D17), a
		ld	l, (ix+0Dh)
		ld	h, (ix+0Eh)
		ld	(word_5D18), hl
		ld	hl, unk_5D16
		call	dsktra
		jp	z, seldsk23
		pop	hl
		ld	hl, 0
		ret

seldsk1a:	call	dbtrw			;Puffer freimachen und auf Lesen schalten
		ld	hl, dbnb
		ld	(hl), 0FFh		;Puffernummer ist ungueltig, obwohl evtl.
						;alle anderen Pufferparameter stimmen
						;(da nur 128 Bytes vom Puffer belegt sind)
		ld	a, c
		ld	(dfrmdv), a
		ld	(dbdev), a
		xor	a
		ld	(ix+0Fh), a		;ruecksetzen alle Flags
		ld	(ix+13h), a		;keine Weiter-Numm. auf Ruecks.
		ld	(ix+0Dh), a		;keine Systemspuren
		ld	(dfrmtr), a		;Spur 0
		ld	(dbtrk), a
		ld	(dbtrk+1), a
		inc	a
		ld	(dbsec), a		;Sektor 1
		ld	(dbsnb), a		;1 phys. Sektor lesen
; Stellen  Sektor-Translate-Tabelle 1,2,3,...
		push	ix
		ld	b, 1Ah			;max. Laenge der Tabelle
seldsk2:	ld	(ix+19h), a
		inc	ix
		inc	a
		djnz	seldsk2
		pop	ix
; Positionieren auf Spur 0 und analysieren Systemspuren
		call	dsidtr			;beliebigen Sekt.Id lesen
		jr	nz, seldsk6		;Fehler beim Lesen; Systemsp. annehmen
		ld	a, h
		ld	(dbslc), a		;setzen Spurformat fuer Sp. 0
		ld	hl, dbflg
		ld	b, (hl)			;merken Fehlerprotokoll-Bit
		set	4, (hl)			;kein Fehlerprotokoll
		push	hl
		call	dbtran			;Lesen Spur 0, Sektor 1
		pop	hl
		ld	(hl), b			;setzen Standard fuer Fehlerprotokoll-Bit
		jr	nz, seldsk6		;Fehler, Systemspuren annehmen
		push	iy
		ld	iy, (dbdma)
		ld	a, (iy+0)		;Format-Loeschbyte
		cp	0E5h			;leere Diskette/geloeschter Eintrag?
		jr	nz, seldsk3		;nein, weder/noch
		cp	(iy+1)			;auch danach 0e5h?
		jr	nz, seldsk4		;nein, geloeschter Dir-Eintrag
		dec	(ix+0Dh)		;Flag: leere Disk oder leere Systemsp.
		jr	seldsk4			;dpbofs auf 254
;
seldsk3:	cp	40h			;IBM-Format (SIOS-Daten-Diskette)?
		jr	z, seldsk5		;ja, 0 Systemspuren annehmen
		cp	20h			;<=31 ? ("S"YL ist groesser!)
		jr	nc, seldsk5		;nein, kann kein Directory sein; 0 Systemsp.
;Systemlader fuer PC1715 beginnt mit 02 oder 03
		ld	a, (iy+20h)		;bei SCP1715-Systemdiskette dort 00h, bei MicroDos dort 10h
		or	(iy+21h)		;bei SCP1715-Systemdiskette dort 00h, bei CP/A dort Dir-Eintrag (E5 oder Filename)
		jr	z, seldsk5		;beides 00h, kein Directory; 0 Systemsp.
seldsk4:	dec	(ix+0Dh)		;=255, wenn Directory; =254 wenn vorn leer
seldsk5:	pop	iy

seldsk6:
; Analysieren Datenspur

; SS/DS-Analyse
		ld	(ix+12h), 2		;Doppel-Step-Impulse
		ld	a, 3			;auf LOGIN-Spur
		call	dsidtt			;belieb. SektId Vorderseite lesen
		jr	nz, seldsk9		;Fehler, unsicheres Format
		ld	d, h			;d:=Sektorlcode, e:=Spur

;if dskds
		bit	0, (ix+18h)		;ist es ein DS-Laufwerk?
		jr	z, seldsk7		;nein, SS
		push	de			;merken Sektorlaengencode und Spur
		set	5, (ix+0Fh)
		ld	a, 7			;Rueckseite der Spur, ab der Format bei SS konstant
		call	dsidtt			;beliebigen Sekt.Id Ruecks. lesen
		ld	a, d			;merken side Ruecks.
		ld	l, e			;h:=Sektorlcode, l:=Spur Rueckseite
		pop	de			;wiederherstellen Vorderseite
		jr	nz, seldsk7		;Fehler beim Lesen, SS
		or	a
		sbc	hl, de			;Vorder- gleich Rueckseite?
		jr	z, seldsk8		;ja; 40 Tr/ 80 Tr unterscheiden
seldsk7:	res	5, (ix+0Fh)

; 40/80 Track Analyse
seldsk8:	ld	a, e			;trk
		sub	3			;waren 2 Steps richtig?
		jr	z, seldsk10		;ja
		dec	(ix+12h)		;Einzelstep-Impulse
		sub	3			;waere 1 Step richtig?
		jr	z, seldsk10		;ja
						;nein, unsicher
seldsk9:	ld	(ix+10h), 0FFh		;provozieren 'BAD SECTOR'
		ld	hl, 0			;erzeugen 'SELECT' Error
		ex	(sp), hl
		jp	seldsk23

; DS/SS und 40/80 ist unterschieden, es fehlt Sektorlaenge
seldsk10:	ld	a, d
		ld	(ix+10h), a		;definieren Sektorlaengencode
		or	a			;Sektorlaenge 128?
		jr	nz, seldsk11		;nein, Puffer reicht nicht

; Stellen Sektor-Translate-Tabelle 1,7,13,..
seldsv:		push	ix
		pop	hl
		ld	bc, 19h
		add	hl, bc
		ex	de, hl
		ld	hl, xlt
		ld	c, 1Ah
		ldir

seldsk11:	call	dtrsla			;stellen hl entspr. Sektorlaenge
		jr	nz, seldsk9		;unzulaessig
		ld	bc, 5			;Laenge eines Eintrags
		bit	1, (ix+12h)		;40 Tr. auf 80er LW?
		jr	nz, seldsk12		;ja
		ld	a, (ix+16h)		;40er LW?
		cp	40+1
		jr	c, seldsk12		;ja, 40 Tr. auf 40er LW
		add	hl, bc			;auf 80er Format stellen
seldsk12:	bit	5, (ix+0Fh)		;DS?
		jr	z, seldsk13		;nein
		add	hl, bc			;auf DS Format
		add	hl, bc

; DPB modifizieren entsprechend erkanntem Format
seldsk13:	ld	c, (hl)			;Zahl der benutzten Spuren
		ld	(ix+14h), c
		inc	hl
		ld	c, (hl)			;Anzahl 128er Sektoren pro Spur
		ld	(ix+0),	c
		inc	hl
		ld	c, (hl)			;Anzahl der Dir-Eintraege -1
		inc	hl
		ld	a, (hl)			;Anzahl der Systemspuren
		srl	a			;feste Anzahl erzwungen?
		jr	c, seldsk15		;ja
		bit	7, (ix+0Dh)		;kann Spur 0 Directory sein?
		jr	z, seldsk15		;nein, Standardanzahl setzen
		inc	(ix+0Dh)		;evtl. leere Systemspur?
		jr	z, seldsk14		;nein, 0 Systemspuren
		ld	b, a			;retten Anzahl Systemspuren
		ld	(dbtrk), a		;einlesen evtl. moeglicher Datenbeginn
		push	hl
		call	dbtran			; dauert bei falscher Spur etwas laenger
		pop	hl
		ld	a, b
		jr	nz, seldsk15		;Fehler, mit Systemspuren annehmen
		push	iy
		ld	iy, (dbdma)
		ld	a, (iy+0Eh)		;falls dort Dir, so Byte 14 von Dir =0
		cp	0E5h			;Daten auch leer?
		pop	iy
		ld	a, b
		jr	nz, seldsk15		;nein, vorn liegen nichtbelegte Systemspuren
seldsk14:	xor	a			;sonst ohne Systemspuren
seldsk15:	ld	(ix+0Dh), a
		or	a			;0 Systemspuren?
		jr	z, seldsk16		;ja
		ld	a, c			;Anzahl Dir-Eintraege -1
		cp	192-1			;>=192 Dir-Eintraege ?
		jr	c, seldsk16		;nein
		ld	c, 128-1		;780k hat 128 Dir-Eintraege
seldsk16:	ld	(ix+7),	c
		inc	hl
		ld	c, (hl)			;Abstand der Blockgroessentab.
		ld	b, 0
		add	hl, bc			;auf Blockgroessentabelle
		ld	c, (hl)
		ld	(ix+2),	c
		inc	hl
		ld	c, (hl)
		ld	(ix+3),	c
		inc	hl
		ld	c, (hl)
		ld	(ix+4),	c

; Bestimmen Zahl der abweichenden 128er-Sektoren ab Spur 0
		xor	a
		ld	(dfrmtr), a		;ab Spur 0
						;(gleichz. Spurkorr. bei falschen Doppelstepimpulsen)
seldsk17:	call	dsidtr			;beliebigen SektId lesen
		jr	nz, seldsk18		;Fehler
		ld	a, h			;Sektorlaenge
		cp	(ix+10h)		;gleich Disketten-Rest?
		jr	z, seldsk18		;ja
		ld	a, (ix+0Fh)
		and	3			;bisherige Anzahl abweichender Spuren
		cp	3			;Zaehler voll?
		jr	z, seldsk18		;ja
		inc	(ix+0Fh)		;erhoehen abweich. Spurzahl
		ld	hl, dfrmtr		;naechste (logische!) Spur
		inc	(hl)
		jr	seldsk17

seldsk18:
; Setzen Anzahl der logischen 128er Rekords im Puffer -1
; Es wird immer das Maximum gesetzt, dies setzt eine dichte
; Sektorfolge beim Transfer voraus! (so in dpbstr)
		ld	(ix+11h), 7

; Berechnen Speicherkapazitaet-1 in BDOS-Bloecken aus
; ((Tracks-dpbofs)*dpbspt/(2**dpbbls))-1
		ld	l, (ix+14h)
		xor	a
		ld	h, a			;hl:=log. Spurzahl
		ld	e, (ix+0Dh)
		ld	d, a			;de:=log. offset-Spurzahl
		sbc	hl, de
		ex	de, hl			;de:=log. Daten-Spurzahl
		ld	l, a			;hl:=0
		ld	b, (ix+0)
seldsk19:	add	hl, de
		djnz	seldsk19		;hl:=(tracks-dpbofs)*dpbspt
		ld	b, (ix+2)
seldsk20:	srl	h
		rr	l
		djnz	seldsk20
		dec	hl
		ld	(ix+5),	l
		ld	(ix+6),	h
		ld	a, (ix+7)		;Dir-Groesse
		inc	a
		and	0FCh
		rrca
		rrca				;div 4, da 4 Dir-Eintr. /Sekt.
		ld	(ix+0Bh), a		;ist Sektorzahl = Check size
		ld	b, (ix+2)		;block shift
seldsk21:	rrca				;BDOS-Bloecke fuer Dir.
		djnz	seldsk21
		ld	b, a
		xor	a
seldsk22:	scf				;Allocation-Bits fuer Dir.
		rra
		djnz	seldsk22
		ld	(ix+9),	a

seldsk23:	pop	hl			;hl auf DPH
		ret				;Return Seldsk

;-----------------------------------------------------------------------------
; home
; auf Spur 0 zurueck (vor jedem Dir-Zugriff)
;-----------------------------------------------------------------------------

home:		ld	hl, dbflg
		bit	2, (hl)			;veraenderter Puffer?
		jr	nz, home1		;ja, nicht freigeben
		ld	a, 0FFh			;sonst Diskwechsel erlauben
		ld	(dbdev), a		;Puffer ist nicht aktiv

;-----------------------------------------------------------------------------
; settrk
; Einstellen Spur in Reg. BC
;-----------------------------------------------------------------------------

home1:		ld	bc, 0
settrk:		ld	(dtrack), bc
		ret

;-----------------------------------------------------------------------------
; Einstellen Sektor in Reg. C
;-----------------------------------------------------------------------------
setsec:		ld	a, c
		ld	(dsectr), a
		ret

;-----------------------------------------------------------------------------
; Einstellen DMA in Reg. BC
;-----------------------------------------------------------------------------
setdma:		ld	(ddma), bc
		ret

;-----------------------------------------------------------------------------
; Uebersetzung Sektornummer
; Translate-Tab-Adr. in DE, Eingangs-Sektornummer in BC,
;			    Ausgangs-Sektornummer in HL
; Es wird keine Translate-Tabelle benutzt, da die Sektor-
; nummernverwaltung verallgemeinert im nicht-Standard-DPB
; enthalten ist (auch fuer physische Sektorlaenge <>128)
;-----------------------------------------------------------------------------
sectran:	ld	h, b
		ld	l, c
		inc	hl		;Sektoren zaehlen in CP/A ab 1
		ret

; Ermitteln DPH und DPB von (c) nach HL bzw. IX
; ret z		DPH existiert nicht, HL=0
; ret nz	ok, HL auf DPH, IX auf DPB
; BC,DE bleibt erhalten
dgetpb:		push	de
		ld	hl, 0
		ld	a, c

		IF RAMDSK=1
		cp	3		; nur LW 0..2 erlaubt
		ELSE
		cp	2		; nur LW 0..1 erlaubt
		ENDIF

		jr	nc, dgetpb1 	; sonst Abbruch
		ex	de, hl		; DE = 0
		ld	hl, dphtab	; Adresstabelle fuer DPH's
		ld	e, c
		add	hl, de
		add	hl, de
		ld	a, (hl)
		inc	hl
		ld	h, (hl)
		ld	l, a		; HL=Adr. zum LW gehriger DPH
		push	hl
		ld	e, 10
		add	hl, de
		ld	e, (hl)
		inc	hl
		ld	d, (hl)		; de:=^dpb
		push	de
		pop	ix		; IX=Adr. zugehriger DPB
		pop	hl
dgetpb1:	pop	de
		ld	a, h
		or	l		; ret z bei hl=0
		ret

;-----------------------------------------------------------------------------
; Schreiben Sektor
; Register C (vom BDOS gestellt):
;	=0, wenn normales write
;	=1, wenn directory-write (sofort ausgeben)
;	=2, wenn Beginn eines neuen Datenblocks (kein preread)
;-----------------------------------------------------------------------------
write:		ld	e, c			;retten Reg. C
		ld	a, (ddrive)
		ld	c, a
		call	dgetpb			;ix:=dpb(ddrive)
		jr	z, drwerr		;Geraet ex. nicht

		ld	hl, dflg
		set	2, (hl)			;Write-Flag setzen
		ld	hl, dbflg
		set	0, (hl)			;Annahme preread notwendig
		ld	a, e			;A:= Write-Typ
		ld	(dwrtyp), a		;merken Write-Typ
		cp	2			;write to unallocated?
		jr	nz, write1		;nein
		ld	a, (ix+3)		;Zahl der 128-Sekt. im Block -1
		inc	a
		ld	(unacnt), a		;fuer diese Sekt. kein preread
		ld	hl, unadev
		ld	(hl), c
		ld	hl, (dtrack)
		ld	(unatrk), hl
		ld	a, (dsectr)
		ld	(dbnb), a

write1:		ld	hl, unacnt
		ld	a, (hl)
		or	a			;noch nicht geschr. Sekt. da?
		jr	z, alloc		;nein
		dec	a			;sonst Restzahl -1
		ld	(hl), a
		inc	hl
		ld	a, c
		cp	(hl)			;gleiches Geraet?
		jr	nz, alloc		;nein
		inc	hl
		ld	a, (dtrack)
		cp	(hl)			;gleiche Spur?
		jr	nz, alloc		;nein

		inc	hl
		ld	a, (dtrack+1)
		cp	(hl)
		jr	nz, alloc

		inc	hl
		ld	a, (dsectr)
		cp	(hl)			;gleicher Sektor?
		jr	nz, alloc		;nein
; Vorbereiten naechsten unalloc write-Aufruf
		cp	(ix+0)			;neue Spur?
		jr	c, write2		;nein

		dec	hl
		ld	d, (hl)
		dec	hl
		ld	e, (hl)
		inc	de
		ld	(hl), e
		inc	hl
		ld	(hl), d
		inc	hl
		xor	a			;Sektor auf Spuranfang
write2:		inc	a			;naechster Sektor
		ld	(hl), a
		ld	hl, dbflg
		res	0, (hl)			;anzeigen kein preread notw.
		jr	drw

; Abbruch read/write mit E/A-Fehler
drwerr:		ld	a, 1
		ret

;-----------------------------------------------------------------------------
; Lesen Sektor
;-----------------------------------------------------------------------------
read:		ld	a, (ddrive)
		ld	c, a
		call	dgetpb			;ix:=dpb(ddrive)
		jr	z, drwerr		;Geraet ex. nicht
		ld	hl, dflg
		res	2, (hl)			;Lesen anzeigen
		ld	hl, dbflg
		set	0, (hl)			;preread notwendig
		ld	a, 2
		ld	(dwrtyp), a

alloc:		xor	a
		ld	(unacnt), a		;Ende unalloc

; gemeinsamer Zweig read/write Floppy
drw:		ld	a, (dsectr)		;Sektornummer in 1..dpbspt ?
		dec	a			;ab 0 zaehlen
		jp	m, drwerr		;<0, Fehler
		cp	(ix+0)			;<dpbspt?
		jr	nc, drwerr		;nein, Fehler
		ld	b, a			;retten dsectr-1
		ld	a, (ix+0Fh)
		and	3			;Anzahl der abweich. 128er Spuren

		ld	e, a
		ld	d, 0
		ld	hl, (dtrack)
		or	a
		sbc	hl, de

		jp	c, drwsec		;ja, Sektorlaenge 128

		ld	a, (ix+10h)		;phys. Sektorlaenge 128?
		or	a			;d.h. mit Sektorversatz?
		jp	z, drwsec		;ja, keine Pufferung sinnvoll

		ld	a, b
		ld	d, (ix+11h)		;Puffermaske
		and	d			;rel. Sekt.nr. des BDOS-Blocks zum Puff.anf.
		push	af			;merken fuer move
		xor	b			;a ist Sektornr. des Puff.anf.
drwbnr:		srl	a			;ermitteln Puffernr in Spur
		rr	d			;noch Bits in Puffermaske?
		jr	nz, drwbnr		;ja
; Test, ob dieser Puffer da ist
		push	af			;merken Puffernummer
		ld	hl, dbnb
		cp	(hl)			;richtige Puffernummer?
		jr	nz, dbn			;nein

		ld	hl, (dtrack)
		ld	de, (dbtrk)
		or	a
		sbc	hl, de
		jr	nz, dbn

		ld	a, (dbdev)		;zum Puffer gehoeriges Geraet
		cp	c			;richtiges Geraet?
		jr	z, dbmat		;ja, Sektor steht im Puffer

dbn:		push	bc			;retten ddrive
		call	dbtrw			;veraenderten Puffer ausgeben
		pop	bc			;c:=ddrive
		ld	hl, dbdev
		ld	(hl), c

		ld	hl, (dtrack)
		ld	(dbtrk), hl

		call	dgetpb			;ix:=dpb(c)
		ld	b, (ix+10h)
		ld	hl, dbslc
		ld	(hl), b
		ld	a, (ix+11h)		;Zahl der 128er Sektoren -1
		inc	a
		add	a, a			;CY:=0
		inc	b			;fuer djnz, falls dpbslc=0
dbsnbz:		rrca				;Zahl der phys. Sekt. im Puff.
		djnz	dbsnbz
		ld	d, a			;merken
		xor	a
		pop	bc			;b:=geforderte Puffernr.
		push	bc
		ld	hl, dbnb
		ld	(hl), b
		inc	b
		dec	b			;Puffernr. 0?
		jr	z, dbsec0		;ja
dbsecz:		add	a, d			;0. phys. Sekt. im Puffer
		djnz	dbsecz			;=dbsnb.*Puffernr.
dbsec0:		ld	e, a			;merken Pufferanf.-Sektor
		ld	b, (ix+10h)
		ld	a, (ix+0)
		inc	b			;wegen djnz
		add	a, a
dbsecm:		srl	a			;phys. Sektoranzahl
		djnz	dbsecm
		sub	e			;Restsektorzahl auf Spur
		cp	d			;>= Zahl zu transferierender?
		jr	nc, dbseco		;ja
		ld	d, a			;sonst nur Restsektorzahl
dbseco:		inc	e			;Sektoren zaehlen ab 1
		ld	a, e
		ld	(dbsec), a		;Sektor
		ld	a, d
		ld	(dbsnb), a		;Sektorzahl
		ld	hl, dbflg
		res	2, (hl)
		bit	0, (hl)			;muss neuer Puffer gel. werden?
		jr	nz, dbprr		;ja
; nur dann kein preread, wenn Rest BDOS-Block >= phys. Puff.
; und Rest BDOS-Block auf Pufferanfang beginnt
		ld	a, (unacnt)
		cp	(ix+11h)		;Restzahl-1 >= Puffergr.-1 ?
		jr	c, dbprr		;nein, preread
		ld	a, (dsectr)		;Sektornr.
		dec	a			;ab 0 zaehlen
		and	(ix+11h)		;rel. Sekt.nr. des BDOS-Blocks zum Pufferanfang
		jr	nz, dbprr
		ld	(dberrf), a		;fehlerfrei gel. anzeigen
dbprr:		call	nz, dbtran		;Puffer lesen

; Sektor in/aus Puffer holen
dbmat:		pop	af			;Puffernr. in Spur wegschm.
		pop	af			;128-er Index im Puffer
		push	af
		rra				;Sektornb*128 berechnen
		ld	b, a
		ld	c, 0
		rr	c
		ld	hl, fdcbuffer
		add	hl, bc			;Adresse im Puffer
		;;ld	de, rwbuffer	;--> in CPA: ld	de,(ddma)	;Nutzer-DMA
		ld	de,(ddma)	;Nutzer-DMA
		ld	a, (dflg)
		bit	2, a			;Schreiben?
		jr	z, dsmove		;nein
		ld	a, (dbflg)
		set	2, a			;anzeigen Puffer beschrieben
		ld	(dbflg), a
		ex	de, hl
dsmove:		ld	bc, 128
		call	flldir
		pop	af			;128-Index im Puffer
		cp	(ix+11h)		;letzter Sektor im Puffer
		jr	z, dmovew		;ja, Puffer ausgeben
		ld	a, (dwrtyp)
		cp	1			;write to directory?
dmovew:		call	z, dbtrw		;ja, veraend. Puffer ausgeben
		ld	a, (dberrf)		;Ergebnis letztes dbtran
		ret

; Schreiben Puffer, wenn notw.
dbtrw:		ld	hl, dbflg
		bit	2, (hl)			;Puffer veraendert?
		ret	z			;nein, Schreiben unterdruecken

; gemeinsamer Zweig Lesen/Schreiben Puffer
dbtran:		ld	hl, dbflg
		call	dsktra
		ld	(dberrf), a		;Fehlerflag stellen
		ld	hl, dbflg
		res	2, (hl)			;Puffer ist nicht beschrieben
		ret

; Lesen beliebigen Sektor-Id.
dsidtt:		ld	(dfrmtr), a		;Eingang fuer Spur in A
dsidtr:		ld	hl, dfrcdb
		jr	dsktra

; gemeinsamer Zweig Read/Write 128-Sektor einzeln
drwsec:		ld	hl, dflg


;*********************************************************
; Umkleidung des physischen Disketten-Transfers
;*********************************************************

; Parameter: HL auf CDB mit folgender Struktur:
; +0: cdbfl	;Flags	(Struktur angepasst an ft.kom)
diof00	equ	0	;** frei fuer Anw. ** (frueher Verify nach Schreiben)
diofid	equ	1	;fest	;=1, wenn nur Sektid. zu lesen
diofwr	equ	2	;fest	;=1, wenn Schreiben
dioffm	equ	3	;fest	;=1, wenn temporaer FM-Format erzwungen
dioftr	equ	4		;=1, wenn keine Fehlermeldung (und -behandlg)
diofhd	equ	5		;=1, wenn Kopf hochnehmen ("noerr")
diofps	equ	6		;=1, wenn trk,sid,sec schon physisch
diofs1	equ	7	;fest	;=1, wenn Rueckseite

; ab hier unwichtig bei "diofhd" in cdbfl =1 (noerr)
; +1: cdbdev	;logisches Geraet 0 .. dphnb-1
; +2: cdbtrk	;log./phys. Spur
; +3: cdbsid	;bel./phys. Seite
; +4: cdbsec	;log./phys. Sektor
; +5: cdbslc	;Sektorlaengencode (0=128, 1=256, 2=512, 3=1024)
; +6: cdbsnb	;Anz. zu uebertr. ph. Sekt. (wenn =0, so nur position.)
; ab hier nur wichtig, wenn "diofid" in cdbfl =0:
; +7,8: cdbdma	;Transferadresse

; Return, falls nicht "diofhd" in cdbfl =1:
;	A=0 (ret z) bei fehlerfrei, sonst A=1 (ret nz) bei cdbfl, "dioftr" =0
;	oder Returncode phys. Transfer
;	E:=trk, D:=sid ,L:=sec, H:=len
;	BC,IX unveraendert

dsktra:		push	bc
		ld	de, diocdb		;CDB des Nutzers auf internen Hilfsspeicher
		ld	bc, 9			;da modifiziert
		push	de
		call	flldir			;falls Original im ROM-Adressbereich
		pop	hl			;hl auf diocdb
		pop	bc
		bit	5, (hl)			;Kopf hochnehmen?
		jp	nz, noerr		;ja, kein Transfer
		push	bc
		push	ix
		push	hl
		inc	hl
		ld	c, (hl)			;logisches Laufwerk
		call	dgetpb			;IX auf DPB stellen
		pop	hl			;wiederherstellen hl
		ld	a, 'D'			;falls LW nicht ex.
		jp	z, dioert		;LW ex. nicht
		bit	6, (hl)			;trk,sid,sec schon physisch?
		jr	nz, diophy		;ja
; Track, Side und Sector entpsr. Diskettenformat setzen
		inc	hl
		inc	hl
		ld	d, (hl)			;Track
		ld	e, d			;merken
		inc	hl
		inc	hl
		ld	c, (hl)			;Sektornummer ab 1
		dec	c			;Sektornr. ab 0
		bit	4, (ix+0Fh)		;Fortsetzung Dsk. auf Ruecks.?
		jr	nz, diots2		;ja
		bit	5, (ix+0Fh)		;ungerade Spuren auf Rueckseite?
		jr	z, diotrs		;nein, einseitig
		srl	d			;Spur halbieren
		jr	nc, diotrs		;gerade Spur, Vorderseite
		jr	diots1			;auf Rueckseite
diots2:		ld	b, (ix+14h)		;log. Spurzahl
		srl	b			;Spuren auf Vorderseite
		ld	a, d
		sub	b			;Spur auf Vorderseite?
		jr	c, diotrs		;ja, Spur in d unveraendert lassen
						;40 ->  0; 41 ->  1; ...; 79 ->  39
						;77 ->  0; 78 ->  1; ...;153 ->  76
		bit	3, (ix+0Fh)		;Ruecks. von aussen nach innen?
		jr	nz, diots4		;ja
		cpl				;40 -> -1; 41 -> -2; ...; 79 -> -40
						;77 -> -1; 78 -> -2; ...;153 -> -77
		add	a, b			;40 -> 39; 41 -> 38; ...; 79 ->   0
						;77 -> 76; 78 -> 75; ...;153 ->   0
diots4:		ld	d, a
diots1:		ld	a, (diocdb)
		set	7, a			;vermerken Rueckseite
		ld	(diocdb), a
		ld	a, (ix+13h)		;Versch. der Sektornr. auf Rueckseite
		add	a, c			;Sektornr. evtl. weiterzaehlen
		ld	c, a
		ld	a, 1			;side:=1
		jr	diotss

diotrs:		xor	a			;side:=0
diotss:		ld	(diocsi), a		;phys. Seite setzen
		ld	a, d
		ld	(dioctr), a		;phys. Spur setzen
		ld	a, e			;log. Spurnr. zurueck nach A
		push	ix
		pop	hl
		ld	de, 19h			;d:=0
		add	hl, de
		cp	(ix+0Dh)		;Systemspur?
		jr	c, diossp		;ja, Sektornummern nicht logisch
		ld	e, c
		add	hl, de			;HL entspr. Index in dpbstr
		ld	c, d			;c:=0
diossp:		ld	a, c			;phys. Sektornr. ab 0
		add	a, (hl)			;+echte Sektornr.
		ld	(diocse), a		;phys. Sektor setzen
diophy:
; i ret, BC, IX im Stack
; i cdb auf diocdb
; i IX auf DPB

;interne Wiederholung dsktra
		push	ix			;retten IX fuer evtl. Wiederholung
		ld	hl, diocdb
		push	hl
		ld	a, (hl)			;side 0/1, -/FM, read/write, -/sectid l.
		and	8Eh			;(1 shl diofs1)+(1 shl dioffm)+(1 shl diofwr)+(1 shl diofid)
		ld	b, a
		ld	a, (ix+18h)		;5"/8", FM/MFM, 40/80, Verify nach Schreiben
		and	78h			;and (1 shl dpbt80)+(1 shl dpbt5z)+(1 shl dpbtdd)+(1 shl dpbtwv) ;aus DPB
		xor	b
		bit	4, (hl)			;Fehlerbehandlung unterdruecken?
		jr	z, dionft		;nein
		inc	a
dionft:		ld	(ft.kom), a
		inc	hl
		ld	a, (ix+15h)		;physische LW-Nr.
		ld	(ft.lwn), a
		inc	hl
		ld	a, (hl)
		ld	(ft.trk), a
		inc	hl
		ld	a, (hl)
		ld	(ft.sid), a
		inc	hl
		ld	a, (hl)
		ld	(ft.sec), a
		inc	hl
		ld	a, (hl)
		ld	(ft.len), a
		inc	hl
		ld	a, (hl)
		ld	(ft.anz), a		;Anzahl der Stepimpulse
		ld	a, (ix+12h)		;Schrittzeit
		ld	(ft.stp), a
		ld	a, (ix+17h)
		ld	(ft.sti), a
		ld	hl, (diocad)
		ld	(ft.adr), hl
		call	floppy			;$$$ phys. Transfer $$$
		pop	hl			;HL wieder auf Paramfeld-Adresse
		pop	ix			;IX wieder auf DPB

; Fehlerprotokoll
dioert:		bit	4, (hl)			;Fehlerprotokoll und -behandlung  unterdr.?
		jr	nz, dtrok		;ja
		or	a			;Returncode
		jr	z, dtrok		; -> fehlerfrei
		cp	'R'			;not ready?
		jp	z, diophy		;ja, Wiederholung
		ld	(derrcd), a		;Fehlercode
		bit	2, (hl)			;war lesen?
		ld	a, 'R'
		jr	z, derrr		;ja
		ld	a, 'W'
derrr:		ld	(derrrw), a
		ld	hl, ft.trk		;phys. Spur
		ld	de, derrtr	; "XXXXXX\a"
		call	mbreco		; (HL) nach hex	konvertieren, Eintragen	nach (DE), 2x inc DE
		ld	hl, ft.trk+1
		call	mbreco		; (HL) nach hex	konvertieren, Eintragen	nach (DE), 2x inc DE
		ld	hl, ft.sec		;Seite
		call	mbreco		; (HL) nach hex	konvertieren, Eintragen	nach (DE), 2x inc DE
		ld	hl, derrms
		call	biosms
		ld	a, 1		;return mit Fehler
dtrok:		or	a		;stellen Z-Flag
		ld	de, (ft.trk)	;E:=trk, D:=sid
		ld	hl, (ft.sec)	;L:=sec, H:=len
		pop	ix
		pop	bc
		ret

; Stellen HL entsprechend Sektorlaengencode in (A)
; ret nz bei unzulaessigem
dtrsla:		or	a
		ld	hl, dtrsl0	; bei A=0 (128)
		ret	z
		dec	a
		ld	hl, dtrsl1	; bei A=1 (256)
		ret	z
		dec	a
		ld	hl, dtrsl2	; bei A=2 (512)
		ret	z
		dec	a
		ld	hl, dtrsl3	; bei A=3 (1024)
		ret

;************************************************
;	Steuertabellen
;************************************************

; Modifizierungstabellen entspr. Sektorlaenge und LW-Typ
;-------------------------------------------------------

; Struktur:

dsltrk	equ	0		;benutzte Spuren
dslspt	equ	dsltrk+1	;Sektoren/Spur
dsldir	equ	dslspt+1	;Dir-Eintraege
dsloff	equ	dsldir+1	;2*offset
				;[+Flag fuer festes offset]
dslfo	equ	1		;festes offset
dslvo	equ	0		;offset =0, falls Directory mgl.
dslblk	equ	dsloff+1	;rel. Adr. Tab. BDOS-Blockgroesse
dsll	equ	5		;Laenge eines Eintrags

; Reihenfolge
; 5", 40 Tr, SS
; 5", 80 Tr, SS
; 5", 40 Tr, DS
; 5", 80 Tr, DS

; 128
dtrsl0:
	db	40,26,63,2*2+dslvo,dbl1k-$-4	;CP/M Standard
	db	80,26,127,2*2+dslvo,dbl2k-$-4
	db	80,26,127,2*0+dslfo,dbl2k-$-4
	db	160,26,127,2*0+dslfo,dbl2k0-$-4

; 256
dtrsl1:
	db	40,32,63,2*3+dslfo,dbl2k-$-4	;SCP Hausformat A51xx
	db	80,32,63,2*3+dslfo,dbl2k-$-4	;SCP
	db	80,32,127,2*4+dslfo,dbl2k-$-4
	db	160,32,127,2*4+dslfo,dbl2k0-$-4	;SCP Hausformat PC1715

; 512
dtrsl2:
;		BIOSVER = 'CPMZ9OK'
;	db	40,36,63,2*0+dslfo,dbl1k-$-4
;	db	80,36,127,2*0+dslfo,dbl2k-$-4
;	db	80,36,127,2*0+dslfo,dbl2k-$-4
;	db	160,36,127,2*0+dslfo,dbl2k0-$-4

	db	40,36,0,2*0+dslfo,dbl1k-$-4
	db	80,36,0,2*0+dslfo,dbl2k-$-4
	db	80,36,0,2*0+dslfo,dbl2k-$-4
	db	160,36,0,2*0+dslfo,dbl2k0-$-4

; 1024
dtrsl3:
	db	40,40,63,2*2+dslvo,dbl1k-$-4	;CP/A Standard A51xx
	db	80,40,127,2*2+dslvo,dbl2k-$-4	;CP/A
	db	80,40,127,2*0+dslvo,dbl2k-$-4	;CP/A
	db	160,40,191,2*4+dslvo,dbl2k0-$-4	;CP/A (und SCP)

; Modifizierungstabellen fuer BDOS-Blockgroesse
dbl1k:	db	3,7,0		;Bl. Shift, Bl. Mask, Ext. Mask
dbl2k:	db	4,0fh,1
dbl2k0:	db	4,0fh,0		;fuer Kapazitaet >255

; log. Sektortranslate-Tabelle fuer 26*128
xlt:	db	1,7,13,19,25,5,11,17,23,3,9,15,21
	db	2,8,14,20,26,6,12,18,24,4,10,16,22


; Fehlermeldung Disk I/O Error
; Genau so lang (kurz), wie in Statuszeile Platz dafuer da ist!
; und das sind 17 Bytes!

derrms:		db	 0Dh ;
		db	 0Ah ;
derrrw:		db	'X'
derrcd:		db	"X;T,Si,Se="
derrtr:		db	"XX"
derrts:		db	"XX"
derrsc:		db	"XX"
		db	7
		db	0
unk_5D16:	db  10h
byte_5D17:	db 0FFh
word_5D18:	dw 0
		db    1 ;
		db    1 ;
		db    0 ;

; CDB fuer Bestimmung Spurformat
dfrcdb:		db	 12h 		;(1 shl diofid)+(1 shl dioftr)
dfrmdv:		db	0FFh		;Geraet
dfrmtr:		db	0FFh		;Spur
		db	0FFh 		;Seite (bel.)
		db	   1 		;Sektor (bel.)
		db	   1 		;Sektorlaengencode (bel.)
		db	   1 		;Sektoranzahl (>0)

; Puffer-CDB
dbflg:	db	0
dbdev:	db	0FFh
dbtrk:	db	0FFh
dbsid:	db	0FFh
dbsec:	db	0FFh
dbslc:	db	0FFh
dbsnb:	db	0FFh
dbdma:	dw	fdcbuffer

; CDB fuer ungepufferte E/A
dflg:	db	0
ddrive:	db	0
dtrack:	db	0FFh
	db	0FFh 		;Seite
dsectr:	db	0FFh
	db	   0 		;128 Bytes Sektorlaenge
	db	   1 		;immer nur 1 Sektor

unacnt:	db	0
unadev:	db	0FFh
unatrk:	db	0FFh
unasec:	db	0

dbnb:	db	0FFh
dwrtyp:	db	0FFh
dberrf:	db	0		;Ergebnisflag letztes dbtran

;Arbeitsbereiche logischer Floppy-Treiber im BIOS-RAM
flldir:		ldir		;Floppy-Puffer <-> Nutzer-DMA
		ret

;CDB-Hilfsspeicher fuer dsktra-Aufruf
diocdb:	db	0FFh
diocde:	db	0FFh
dioctr:	db	0FFh
diocsi:	db	0FFh
diocse:	db	0FFh
diocsl:	db	0FFh
diocsn:	db	0FFh
diocad:	dw	0FFFFh


;************************************************
; physischer Disketten-Transfer
;************************************************

	include biosfdc.inc

;;;-----------------------------------------------------------------------------
;;; Ausgabe Zeichenkette
;;;-----------------------------------------------------------------------------
;;prnst:		ex	de, hl
;;		ld	hl, (0F057h)	; Adr. PRNST-Funktion des Z1013-OS
;;		jp	(hl)
;;biosms:		di
;;		call	prnst
;;		ei
;;		ret
;;
;;;-----------------------------------------------------------------------------
;;; Kurze Warteschleife
;;;-----------------------------------------------------------------------------
;;wait:		push	bc
;;		ld	c, 0Ah
;;wait1:		ld	b, 0BDh
;;wait2:		djnz	wait2
;;		dec	c
;;		jr	nz, wait1
;;		pop	bc
;;		djnz	wait
;;		ret
;;
;;;-----------------------------------------------------------------------------
;;; (HL) nach hex	konvertieren, Eintragen	nach (DE), 2x inc DE
;;;-----------------------------------------------------------------------------
;;mbreco:		ld	a, (hl)
;;hexa:		call	hexa1		; A nach hex konvertieren, Eintragen nach (DE),	2x inc DE
;;hexa1:		rrca
;;		rrca
;;		rrca
;;		rrca
;;		push	af
;;		and	0Fh
;;		sub	0Ah
;;		jr	c, hexa2
;;		add	a, 7
;;hexa2:		add	a, 3Ah
;;		ld	(de), a
;;		inc	de
;;		pop	af
;;		ret
;;


;;;-----------------------------------------------------------------------------
;;; cboot boot
;;;-----------------------------------------------------------------------------
;;
;;cboot:
;;		jp	0F000h		; Sprung in Z9001-OS (Kaltstart)

;;;-----------------------------------------------------------------------------
;;; CCP von @OS.COM wiederhst.
;;;-----------------------------------------------------------------------------
;;
;;;;wboot:
;;
;;
;;ccplod:
;;	ld	c,13		;;Reset Disksystem (LOGIN neu)
;;	call	bdos
;;	ld	a,0
;;ccpldv	equ	$-1		;;Kaltstart-Laufwerk (A=1,B=2,..)
;;	ld	(fcb_os),a
;;	xor	a
;;	ld	e,a		;;user 0
;;	ld	(fcb_os+12),a	;;extent
;;	inc	a		;;ab Satz 1 lesen
;;	ld	(fcb_os+32),a
;;	ld	c,32		;;set user 0
;;	call	bdos
;;	ld	c,15		;;open
;;	call	de_os
;;	inc	a		;;Datei da?
;;	jr	z,nosys_	;;nein
;;	ld	b,(800h+127)/128	;;128er records Lnge CCP
;;	ld	hl,CCP		;;Zieladresse
;;;;	ld	hl,8000h	; test
;;ccpldz:	push	bc
;;	push	hl
;;	ex	de,hl
;;	ld	c,26		;;setdma
;;	call	bdos
;;	call	rd_os		;;read
;;	or	a		;;Fehler?
;;	jr	nz,nosyst	;;ja
;;
;;	dbg	'H'
;;
;;	ld	hl,(CCP+1)	;;Test, ob richtiges @OS.COM
;;	ld	bc,0		;;bevor BDOS-Anfang ueberladen!
;;ccpjpm	equ	$-2		;;vom Kaltstart hinterlegt
;;	sbc	hl,bc		;;stimmt Sprungadresse?
;;	jr	nz,nosyst	;;nein, falsches CCP
;;	pop	hl
;;	ld	bc,80h
;;	add	hl,bc
;;	pop	bc
;;	djnz	ccpldz
;;;	jr	ccplde		;ok
;;ccplde:	
;;;;	ld	hl, BDOS
;;;;	ld	bc, 5
;;;;	ld	(HL), 0
;;;;	ld	DE, BDOS+1
;;;;	ldir			;BDOS-Kopf wiederherstellen
;;	dbg	'Z'
;;	ret
;;
;;nosyst:	pop	hl
;;	pop	bc
;;nosys_:
;;	dbg	'e'
;;	ld	hl,nosysd
;;	call	biosms
;;	ret		;; test
;;;;	call	headup		;;Verriegelung oeffnen
;;	ld	b,200		;;200* 10ms
;;	call	wait1		;;warten
;;	jr	ccplod		;;neuer Versuch
;;nosysd:
;;	db	0dh,0ah,"No Sysdisk!",0
;;
;;fcb_os:	db	01		;;LW
;;	db	"@CPMZ9  COM"
;;	db	21 dup(0)
;;rd_os:	ld	c,20
;;de_os:	ld	de,fcb_os
;;	jp	bdos


;-----------------------------------------------------------------------------
; wboot,
;-----------------------------------------------------------------------------

wboot:
		di
		;192K-Modul
		ld	a,bank_ccpkopie		; CCP-Kopie
		out	(0ffh),a
		ld	hl, 0e000h
		ld	de, CCP
		ld	bc, 0800h	; Lnge CCP
		ldir

		ld	a,bank_cpm		; Bank FF: CPM einschalten
		out	(0ffh),a

		;sind wir im richtigen Bereich?
		MESSAGE "WBOOT \{$} noch < 0E000?" 

wboot00:
		ld	hl, CCP+0003h

wboot3:		ld	(wboot3u1+1), hl

		ld	sp, 80h

		di
		ld	a, hi(intvectab)	;E7H
		ld	i, a
		im	2

		ld	a, 0C3h		; Befehlscode JP
		ld	(0), a		; 0000h: jp wboot
		ld	(38h), a	; 0038h: RST 38H, jp wboot
		ld	hl, ubios+3	; wboot	im upper Bios
		ld	(1), hl
		ld	(39h), hl
		ld	(5), a		; 0005h: jp bdos
wboot0:		ld	hl, BDOS
		ld	(6), hl

		ld	a, 0C7h
		out	(CTC3), a
		ld	a, 40h
		out	(CTC3), a

		ld	a, 0FFh
		ld	(dbdev), a

wboot1:		ld	hl, 3		; aktuelles Laufwerk
		ld	c, (hl)
		push	bc
		ld	a, c
		and	0Fh		; current user 0 setzen
		ld	c, a
		xor	(hl)
wboot2:		or	0		; Std-Laufwerk, wird von cold gepatcht
		ld	(hl), a
		ld	e, 1
		call	seldsk
		pop	bc
		ld	a, h
		or	l
		jr	z, wboot1	; bei SELDSK-Fehler ist HL=0

;;		;CCP laden
;;		ld	a,(firstrun)
;;		or	a
;;		call	z,ccplod
;;		xor	a
;;		ld	(firstrun),a

		ld	hl, 3		; aktuelles Laufwerk
		ld	c, (hl)
wboot3u1:	jp	CCP+0003h	; CLEAR: CCP-Aufruf mit	Lschen	des Befehlsbuffers

;-----------------------------------------------------------------------------
; write disk
;-----------------------------------------------------------------------------
_write1:	
		IF RAMDSK=1
		ld	a,(ddrive)	; Laufwerk
		cp	2		; RAM-disk C?
		jp	z,WRITE_RD
		ENDIF

		jp	write		; WRITE	fr Floppy

;-----------------------------------------------------------------------------
; read disk
;-----------------------------------------------------------------------------
_read1:
		IF RAMDSK=1
		ld	a,(ddrive)	; Laufwerk
		cp	2		; RAM-disk C?
		jp	z,READ_RD
		ENDIF

		ld	(stackp), sp	; Zwischenspeicher Stackpointer
		ld	sp, biosstack	; Adr. BIOS-Stackbereich
		call	read		; READ im Schattenbios
		ld	sp, (stackp)	; Zwischenspeicher Stackpointer

		or	a
		ret

;-----------------------------------------------------------------------------
; console status
;-----------------------------------------------------------------------------
_const:		call	0F006h		; Z9001-OS: Abfrage Status CONST
		or	a
		ret	z
		ld	a, 0FFh
		ret

;-----------------------------------------------------------------------------
; BAT-Gertereiber fr CRT
;-----------------------------------------------------------------------------
_BATC:		or	a
		jp	nz, 0F7B4h	; Z9001-OS: BAT, Steuerprogramm fr Batch-Mode von CONST
		ld	hl, (0F016h)	; Z9001-OS: Adresse des UP READER (Eingabe Zeichen von Reader)
		inc	hl
		inc	hl
		jp	(hl)

;-----------------------------------------------------------------------------
; Dummy-Gertetreiber, gibt stets Dateiende (^Z) zurck
;-----------------------------------------------------------------------------
_dummyin:				; ^Z; Eingabeende
		ld	a, 1Ah
		ret

;-----------------------------------------------------------------------------
; console character out
;-----------------------------------------------------------------------------
_conout:	ld	a, (byte_0_E34D)
		cp	0FFh
		jr	z, _conout3
		ld	hl, (off_0_E34E)
		res	7, c
		inc	c
		ld	(hl), c
		dec	hl
		ld	(off_0_E34E), hl
		ld	hl, byte_0_E34D
		dec	(hl)
		jr	nz, _conout2
		dec	(hl)
		ld	hl, _conout1
		push	hl
		ld	hl, (0F069h)	; Z9001-OS: Adr. Setzen log. Cursor
		push	hl
		ld	l, 3
		ld	de, (loc_0_E350)
		ret
_conout1:	ld	hl, loc_0_E351
		ld	(off_0_E34E), hl
_conout2:	or	a
		ret
_conout3:	ld	a, (25h)	; KEYBU	(Tastaturbuffer)
		cp	13h		; PAUSE-Taste?
		jr	nz, _conout4
		push	bc
		call	0F009h		; Z9001-OS: Eingabe Zeichen von CONST, PAUSE-Taste holen
		call	0F009h		; Z9001-OS: Eingabe Zeichen von CONST, Warten auf Tastendruck
		pop	bc
_conout4:	ld	a, 1Bh
		cp	c
		jp	nz, 0F00Ch	; Z9001-OS: Ausgabe Zeichen zu CONST
		ld	a, 2
		ld	(byte_0_E34D), a
		jr	_conout1

byte_0_E34D:	db 	0FFh
off_0_E34E:	dw	loc_0_E350
loc_0_E350:	db	0
loc_0_E351:	db	0

;-----------------------------------------------------------------------------
; RAM-Disk-Treiber
; mu nach RAM-Floppy-Bereich C000-DFFF liegen, wg. Bankumschaltung
;-----------------------------------------------------------------------------

		IF RAMDSK=1

READ_RD:	
		if READ_RD < 0e000h 
			WARNING 'READ_RD < 0e000h'
		endif

		CALL	RADR
		JR	C, READ1_RD
		di
		bank	e
		LD	DE,(ddma)
		JR	WRIT1_RD

READ1_RD:	LD	A,1
		RET

WRITE_RD:	CALL	RADR
		JR	C,READ1_RD
		di
		bank	e
		EX	DE,HL
		LD	HL,(ddma)
WRIT1_RD:	
;		bank	bank_rd1
		LD	BC,80H
		LDIR
		bank	bank_cpm
		ei
		XOR	A
		RET

RADR:		LD	H,0
		LD	A,(dtrack)
		LD	L,A	; HL = Tack
		LD	A,(dsectr)
		dec	A	; Sektorenzhlung beginnt mit 1, deshalb Korrektur
		LD	C,A
		LD	B,0

		; HL = Track * Sectors/Track = Basis-Sektor currenttrack
		ADD	HL,HL	; *2
		ADD	HL,HL	; *4
		ADD	HL,HL	; *8 Sectors/Track
		; Sektor offset
		ADD	HL,BC	; B=0; HL = Basis-Sektor currenttrack + currentsector
		; * Sektorlnge
		ADD	HL,HL	; *2
		ADD	HL,HL	; *4
		ADD	HL,HL	; *8
		ADD	HL,HL	; *16
		ADD	HL,HL	; *32
		ADD	HL,HL	; *64
		ADD	HL,HL	; *128 Bytes/Sektor => Offset in RAM-Disk

;		HL < 2000h ? -> Bank 1
;		HL >= 2000h ? -> Bank 2
;		HL >= 4000h ? -> Bank 3

;		ld	a,h
;		cp	40h
;		jr	c, radr2	; wenn a<40
;		sub	a,40h
;		ld	h,a
;		ld	e, bank_rd3
;		jr	radr5
;radr2:		
;		cp	20h
;		jr	c, radr3	; wenn a<20
;		sub	a,20h
;		ld	h,a
;		ld	e, bank_rd2
;		jr	radr5
;radr3:
;		ld	e, bank_rd1


		ld	e, bank_rd1
		ld	a,h
		cp	40h
		jr	c, radr2	; wenn a<40
		ld	e, bank_rd3
		jr	radr5
radr2:		
		cp	20h
		jr	c, radr5	; wenn a<20
		ld	e, bank_rd2

radr5:		and	a, 1fh
		ld	h, a

		LD	BC,0c000h
		ADD	HL,BC	; Basis-Adr. RAM-Disk addieren
		LD	A, 0DFH
		CP	H	; Cy=1: H>0DFH, also RAM-Disk overflow
		RET

		ENDIF


;-----------------------------------------------------------------------------
; Ausgabe Zeichenkette
;-----------------------------------------------------------------------------
prnst:		ex	de, hl
		ld	hl, (0F057h)	; Adr. PRNST-Funktion des Z1013-OS
		jp	(hl)
biosms:		di
		call	prnst
		ei
		ret

;;;-----------------------------------------------------------------------------
;;; Kurze Warteschleife
;;;-----------------------------------------------------------------------------
;;wait:		push	bc
;;		ld	c, 0Ah
;;wait1:		ld	b, 0BDh
;;wait2:		djnz	wait2
;;		dec	c
;;		jr	nz, wait1
;;		pop	bc
;;		djnz	wait
;;		ret

;-----------------------------------------------------------------------------
; (HL) nach hex	konvertieren, Eintragen	nach (DE), 2x inc DE
;-----------------------------------------------------------------------------
mbreco:		ld	a, (hl)
hexa:		call	hexa1		; A nach hex konvertieren, Eintragen nach (DE),	2x inc DE
hexa1:		rrca
		rrca
		rrca
		rrca
		push	af
		and	0Fh
		sub	0Ah
		jr	c, hexa2
		add	a, 7
hexa2:		add	a, 3Ah
		ld	(de), a
		inc	de
		pop	af
		ret



;-----------------------------------------------------------------------------
; CPA-Disk-Parameter
;-----------------------------------------------------------------------------
;Disk-Parameter-Header
dph0:		dw	0		; XLT	translation table
		dw	0
		dw	0
		dw	0
		dw	dirbuf		; DIRBUF
		dw	dpb0		; DPB
		dw	CSV0		; CSV
		dw	ALV0		; ALV

dph1:		dw	0		; XLT	translation table
		dw	0
		dw	0
		dw	0
		dw	dirbuf		; DIRBUF
		dw	dpb1  		; DPB
		dw	CSV1		; CSV
		dw	ALV1		; ALV

		IF RAMDSK=1
;RAMDISK
dph2:		dw	0		; XLT	translation table
		dw	0
		dw	0
		dw	0
		dw	dirbuf		; DIRBUF
		dw	dpb2		; DPB
		dw	0		; CSV
		dw	ALV2		; ALV

		ENDIF

dphtab:		dw	dph0
		dw	dph1

		IF RAMDSK=1
		dw	dph2
		ENDIF

;Disp-Parameter-Blcke
;800K-Laufwerk 80 Spuren DS, DD
dpb0:		dw	40		; SPT	40 Sektoren/Spur
		db 	4		; BSH	= log2(blocksize/128)
					;  => blocksize = (2^4)*128 = 2048
		db 	00001111b	; BLM	= 2^BSH-1
		db 	0		; EXM
		dw	400-1		; DSM	max. 400 Blcke
					;  => also 400*2048 = 800K disc
		dw	192-1		; DRM	max 192 Dir. Eintrge
		db 	11100000b	; AL0	also 3 Blcke (192*32 / 2048)
		db 	0		; AL1	fr DIR reservieren
		dw	48		; CKS	= (DRM+1)/4
		dw	0		; OFF	0 reservierte Spuren
;Erweiternung DBP des Z9001-CBIOS
		db  	00100000b 	; DPB-Steuerinformationen
					; bit 5=1 Double sided zu betreiben
		db    	3 		; Sektorlaengencode
		db    	7 		; Puffermaske 8*128
		db    	1 		; Zahl der Stepimpulse
		db    	0  		; keine Weiter-Numm. auf Rueckseite
		db  	80		; benutzte Zahl phys. Spuren
		db    	0  		; (dbp+15h) phys. LW-Nr. (0 --> A:)
		db  	80		; Gesamtzahl physische Spuren (wird in FORMATZ abgefragt)
		db	4*10		; Schrittzeit fuer 80 Tr. DS LW
		db	00111001b	; vom Laufwerkstyp abhaengige Flags
					; bit 0=0 bei SS, =1 bei DS
					; bit 3=0 bei SD (FM), =1 bei DD (MFM)
					; bit 4=0 bei 8", =1 bei 5"
					; bit 5=0 bei nicht 80-Spur, =1 bei 80-Spur
					; bit 6=0 ohne, =1 mit Verify nach Schreiben
					; bit 7=0 bei max 26 ,=1 bei max 52 Sekt./Spur (8" MFM)
;translation tab dbp0
		db	1,2,3,4,5,6,7,8,9,10,11,12,13,14
		db	15,16,17,18,19,20,21,22,23,24,25,26

dpb1:		dw	40		; SPT	40 Sektoren/Spur
		db 	4               ; BSH	= log2(blocksize/128)
                                        ;  => blocksize = (2^4)*128 = 2048
		db 	00001111b       ; BLM	= 2^BSH-1
		db 	0               ; EXM
		dw	400-1           ; DSM	max. 400 Blcke
                                        ;  => also 400*2048 = 800K disc
		dw	192-1           ; DRM	max 192 Dir. Eintrge
		db 	11100000b       ; AL0	also 3 Blcke (192*32 / 2048)
		db 	0               ; AL1	fr DIR reservieren
		dw	48              ; CKS	= (DRM+1)/4
		dw	0               ; OFF	0 reservierte Spuren
; CPA-Erweiterung
		db  	00100000b 	; DPB-Steuerinformationen
					; Bit5=1 Double sided zu betreiben
		db    	3 		; Sektorlaengencode
		db    	7 		; Puffermaske 8*128
		db    	1 		; Zahl der Stepimpulse
		db    	0  		; keine Weiter-Numm. auf Rueckseite
		db  	80		; benutzte Zahl phys. Spuren
		db    	1  		; (dbp+15h) phys. LW-Nr. (1 --> B:)
		db  	80		; Gesamtzahl physische Spuren (wird in FORMATZ abgefragt)
		db	4*10		; Schrittzeit fuer 80 Tr. DS LW
		db	00111001b	; vom Laufwerkstyp abhaengige Flags
					; bit 0=0 bei SS, =1 bei DS
					; bit 3=0 bei SD (FM), =1 bei DD (MFM)
					; bit 4=0 bei 8", =1 bei 5"
					; bit 5=0 bei nicht 80-Spur, =1 bei 80-Spur
					; bit 6=0 ohne, =1 mit Verify nach Schreiben
					; bit 7=0 bei max 26 ,=1 bei max 52 Sekt./Spur (8" MFM)
; translation tab dbp1
		db	1,2,3,4,5,6,7,8,9,10,11,12,13,14
		db	15,16,17,18,19,20,21,22,23,24,25,26


		IF RAMDSK=1

;DPB RAM-Disk
;dpb2:		Dw	26		; SPT	26 Sektoren/Spur
;		DB	2		; BSH	= log2(blocksize/128)
;;					;  => blocksize = (2^x)*128 =  512
;		DB	3		; BLM	= 2^BSH-1
;		DB	0		; EXM
;		Dw	16-1		; DSM	max. x Blcke
;					;  => also x* blocksize = 8K disc !!!
;		Dw	8-1		; DRM	max x Dir. Eintrge
;		DB	10000000b	; AL0	also x Block (DRM*32/blocksize)
;		DB	0		; AL1	fr DIR reservieren
;		Dw	0		; CKS	= 0 -- fixed
;		Dw	0		; OFF	0 reservierte Spuren

dpb2:		Dw	8		; SPT	26 Sektoren/Spur
		DB	3		; BSH	= log2(blocksize/128)
;					;  => blocksize = (3^x)*128 =  1024
		DB	00000111b	; BLM	= 2^BSH-1
		DB	0		; EXM
		Dw	24-1		; DSM	max. x Blcke
					;  => also x* blocksize = 24K disc !!!
		Dw	32-1		; DRM	max x Dir. Eintrge
		DB	10000000b	; AL0	also x Block (DRM*32/blocksize)
		DB	0		; AL1	fr DIR reservieren
		Dw	0		; CKS	= 0 -- fixed
		Dw	0		; OFF	0 reservierte Spuren

		ENDIF

ubiosend:	equ	$

		dephase

ccpende		equ	$



		phase	loadadress+ccpende-ccpstart

;-----------------------------------------------------------------------------
;EINMALIGE AUSFHRUNG BEIM INITIALEN LADEN
;DIESER CODE WIRD AUF LOADADRESS+OFFSET VERSCHOBEN UND VON DORT GESTARTET
;-----------------------------------------------------------------------------
; cold boot
; Einsprung durch Init nach Bootloader, Initialisierung von cold und ubios
; IN A: Boot-Laufwerk (A=0, B=1, ..)
;-----------------------------------------------------------------------------

cold:		di
		ld	sp, 80h

		;192K-Modul

		push	af		; Boot-Laufwerk sichern
	
		ld	a,bank_ccpkopie		; CCP-Kopie
		out	(0ffh),a
		ld	hl, LOADADRESS
		ld	de, 0E000h
		ld	bc, 0800h	; Lnge CCP
		ldir

		ld	a,bank_cpm		; Bank FF: CPM einschalten
		out	(0ffh),a

		pop	af		; A=Boot-Laufwerk


		; CCP und BDOS und BIOS kopieren
		ld	hl, LOADADRESS
		ld	de, CCP
		ld	bc, ubiosend-CCP	;Lnge CCP+BDOS
		ldir

;-----------------------------------------------------------------------------
;
; startmeldung,
; Initialisieren Gertetreibertabelle,
; RAM-Test,
; starten des CCP-Systems
; Register A: untere 2 Bit Laufwerk; Bit 7 = 1 Warmstart
; Bei Bit7=0 Kaltstart mit kompletter Initialisierung, es
; kann ein Kommando beim start des CCP ausgefhrt werden
;
;-----------------------------------------------------------------------------


;;		xor	a		;Laufwerk 0

		and	3
		ld	(3), a		; Standard-Laufwerk
		ld	(wboot2+1), a

;; f. @OS-Load
;;		inc	a		;von Kaltstart-Laufwerk
;;		ld	(ccpldv),a
;;  		ld	hl,(CCP+1)	;fuer Warmstart-Kontrolle
;;		ld	(ccpjpm),hl	;Sprungadresse aus CCP hinterlegen

		ld	hl, 0FCB0h	; Z9001-OS: Tabelle der Interruptadressen
		ld	bc, 0Ch
		ld	de, intvectab	; nach E700 kopieren
		ldir


		ld	c, stdiobyte	; I/O-Byte: LIST=UL:,PUNCH=UP1:,READER=UR1:,CONST=TTY:
		call	0F03Ch		; Z9001-OS: Setzen I/O-Byte

		ld	a, hi(intvectab); Interrupttabelle
		ld	i, a		; Interruptregister setzen
		im	2

; Initialisierung FDC
		ld	a, 20h		; 0010 0000 FDC	Reset
		out	(FDCZ),	a
		ld	a, 13h		; 0001 0011 Terminal Count aktivieren, Motor LW	1+2 an
		out	(FDCZ),	a
;

binit1:		ld	b, 0		;INITIALISIERUNG P8272
binit2:		djnz	binit2
		in	a, (FDCD)
		cp	80h
		jr	z, binit4
		in	a, (FDCC)
binit3:		jr	binit1
;

stab:
;		BIOSVER = 'CPMZ9OK'
;		db  	10011111b	;XXXX=Schrittratenzeit SRT,XXXX= Kopfladezeit HUT
;		db  	00111111b	;XXXXXXX=Kopfladezeit HLT,X=no dma ND
;		BIOSVER = 'CPMSS4R'
;		db  	11101111b	;XXXX=Schrittratenzeit SRT,XXXX= Kopfladezeit HUT
;		db  	00111111b	;XXXXXXX=Kopfladezeit HLT,X=no dma ND
;		BIOSVER = 'CPMSD8'
;		db 	11101111b
;		db 	11111111b

		db  	10011111b	;XXXX=Schrittratenzeit SRT,XXXX= Kopfladezeit HUT
		db  	00111111b	;XXXXXXX=Kopfladezeit HLT,X=no dma ND
;
binit4:		ld	hl, stab-1	;PARAMETER LADEN
		ld	bc, 303h	;SPECIFY-COMM 3BYTES
		call	wcom1		;SCHREIBEN COMM
		xor	a
		ld	(UNIT), a
		ld	(dFDCZ1), a
		call	recal2
		call	sds
		bit	4, a
		ld	a, 1
		jr	z, binit5
		ld	(dFDCZ1), a
binit5:		ld	(UNIT), a
		call	recal2
		call	sds
		bit	4, a
		ld	a, (dFDCZ1)
		jr	z, binit6
		or	2
		ld	(dFDCZ1), a
binit6:
		out	(FDCZ), a	; ZusatzRegister FDC schreiben

		; Gertetreibertabellen fllen
		di
		ld	hl,fiotab
		ld	de,0EFC9h	; Tabelle der Gertetreiberadressen
		ld	bc,4*4*2
		ldir

		ei

		xor	a
		ld	(lstflag), a

binit7:		ld	hl, loc_0_E350
		ld	(off_0_E34E), hl
		ld	hl, 0
		ld	(loc_0_E350), hl
		ld	a, 0FFh
		ld	(byte_0_E34D), a


		ld	hl, 0f000h	; Patchen BIOS-Verteiler (JP cboot)
		ld	(UBIOS+1), hl


		ld	hl, bootmsg	; Anzeige Boot-Meldung
		call	prnst		; Stringausgabe


		IF RAMDSK=1
		call	RFDEL
		ENDIF

binit12:	ld	hl, startkdo	; start-Kommando
		ld	de, CCP+0007h	; CCP-Befehlsbuffer
		ld	bc, 81h		; 80 Byte + 1 Lngenbyte
		ldir

		ld	hl, CCP		; cboot:	CCP-start mit Ausfhrung des startkdo :-))
		;ld	hl, CCP+0003	; cboot:	CCP-Aufruf mit Lschen des Befehlsbuffers

binit11:	jp	wboot3		; WBOOT	im Bios

;-----------------------------------------------------------------------------
;
;-----------------------------------------------------------------------------

		IF RAMDSK=1

RFDEL:	LD      HL, MT4		; RAMDISK LOESCHEN
        CALL	PRNST
RF1:    CALL	0F009h		; conin im z9001-OS
	and	5FH		; klein -> gro
	CP     'Y'
	JR	Z, RF2
	CP     'N'
	RET	Z        
	JR	NZ, RF1
RF2:
	; Ramdisk mit 0E5h fllen
	bank	bank_rd1
        LD	HL,0c000h
	LD      DE,0c001h
        LD	BC,2000h-1
        LD 	A,0E5H
	LD	(HL),A     
	LDIR

	bank	bank_rd2
        LD	HL,0c000h
	LD      DE,0c001h
        LD	BC,2000h-1
        LD 	A,0E5H
	LD	(HL),A     
	LDIR

	bank	bank_rd3
        LD	HL,0c000h
	LD      DE,0c001h
        LD	BC,2000h-1
        LD 	A,0E5H
	LD	(HL),A     
	LDIR

	bank	bank_cpm
	RET

MT4:	DB	0DH
	DB	0AH
	DB	0AH
	DB	"RAM-Disk formatieren (Y/N)?"
	DB	0

		ENDIF

;-----------------------------------------------------------------------------
;
;-----------------------------------------------------------------------------

fiotab:
		; Gertetreiber	CONST setzen
		dw	uTTYC		; TTY:
		dw	uCRTC		; CRT:
		dw	_BATC		; BAT:
		dw	0F8F1h		; RDR:=	CRT-Treiber Z9001
		; Gertetreibertabelle READER setzen
		dw	0F8F1h		; TTY:=	CRT-Treiber Z9001
		dw	_dummyin	; RDR:
		dw	_dummyin	; UR1:
		dw	_dummyin	; UR2:
		; Gertetreibertabelle PUNCH setzen
		dw	0F8F1h		; TTY:=	CRT-Treiber Z9001
		dw	upunch		; PUN:
		dw	upunch		; UP1:
		dw	uup		; UP2:
		; Gertetreibertabelle LIST setzen
		dw	18h
		dw	0F8F1h		; TTY:=	CRT-Treiber Z9001
		dw	uCRTL		; CRT:
		dw	uup		; LST:
		dw	uUL		; UL:

;-----------------------------------------------------------------------------
;
;-----------------------------------------------------------------------------

;IO-Byte: Bezeichnung phys. Treiber wie bei Z9001 (lt. OS-Handbuch)
;Bit 1,0 (CON:-Kanal):
;     00 (TTY:): Beep + Gross<=>Klein	uTTYC
;     01 (CRT:): nur  Tastatur-Beep	uCRTC
;     10 (BAT:): 			_BATC
;     11 (UC1:): ohne Beep"		CRT-Treiber Z9001
;Bit 3,2 (RDR:-Kanal):
;     00 (TTY:): CRT-Treiber Z9001
;     01 (RDR:): -			_dummyin		SCP: PTR
;     10 (UR1:): -			_dummyin
;     11 (UR2:): -			_dummyin
;Bit 5,4 (PUN:-Kanal):
;     00 (TTY:): CRT-Treiber Z9001
;     01 (PUN:): 			upunch			SCP: PTP
;     10 (UP1:): 			upunch
;     11 (UP2:): 			_up2
;Bit 7,6 (LST:-Kanal):
;     00 (TTY:): CRT-Treiber Z9001
;     01 (CRT:): V24-User-Port 1200 Bd	uCRTL
;     10 (LPT:): V24-User-Port 9600 Bd	uup			Z9001: LST
;     11 (UL1:): Centronics User-Port	uUL

;display I/O-Byte Macro
diob		macro 	shift, compare
		IF stdiobyte & 011b<<(2*shift) = compare<<(2*shift)
			db 	'*'
		ELSE
			db	' '
		ENDIF
		endm

; Sondercodes Z9001:
; 14h nchstes Zeichen ist Code fr Vordergrundfarbe	+ 8 blinken
; 15h nchstes Zeichen ist Code fr Hintergrundfarbe
; 16h Inversdarstellung aller folgenden Zeichen
; Farben: 0 schwarz, 1 rot, 2 grn, 3 gelb, 4 blau, 5 purpur, 6 zyan, 7 wei

bootmsg:	db  	0Ch
		db	14h,1
		db	"CP/A Z9001 fuer 192K-Modul ", DATE, 0Ah, 0Dh
		db	14h,2
		db  	"Schwarzenberg 1989, Pohlers 2007", 0Ah, 0Dh
		db	0Dh, 0Ah
		db	14h,3,"TPA:",14h,2," 100H - \{BDOS-1}H", 0Dh, 0Ah
		db	0Dh, 0Ah
		db 	14h,3
		db	"Laufwerke:", 0Dh, 0Ah
		db	14h,2
		db	"A: Floppy-Disk 5", '"', "(80,DS)", 0Dh, 0Ah
		db	"B: Floppy-Disk 5", '"', "(80,DS)", 0Dh, 0Ah
		IF RAMDSK=1
		db	"C: RAM-Disk 24K", 0Dh, 0Ah
		ENDIF
		db	0Dh, 0Ah
		db 	14h,3
		db	"I/O-Devices:", 0Dh, 0Ah
		db	14h,2
		diob 0, 00b
		db	" CON:=TTY: Beep + Gross<=>Klein", 0Dh, 0Ah
		diob 0, 01b
		db	" CON:=CRT: nur  Tastatur-Beep", 0Dh, 0Ah
		diob 0, 11b
		db	" CON:=UC1: ohne Beep", 0Dh, 0Ah

;		diob 1, 01b
;		db	" RDR:=PTR: dummy (^Z)", 0Dh, 0Ah

		diob 2, 01b
		db	" PUN:=PTP: Ausgabe hexadezimal", 0Dh, 0Ah

		diob 3, 01b
		db	" LST:=CRT: V24-User-Port 1200 Bd", 0Dh, 0Ah
		diob 3, 10b
		db	" LST:=LPT: V24-User-Port 9600 Bd", 0Dh, 0Ah
		diob 3, 11b
		db	" LST:=UL1: Centronics User-Port", 0Dh, 0Ah
		db 	0

; hier kann ein	Befehl eingetragen werden, der beim start des CCP ausgefhrt werden soll
startkdo:	db  	7Fh	; Lnge	des Befehls(buffers)
;		db 	"CLOCK"
		db	128-($-startkdo-1) dup (0)	; mit 00 auffllen



biosend:	equ	$

		dephase

;-----------------------------------------------------------------------------
; BIOS-RAM
;-----------------------------------------------------------------------------

		org	0e200h

; intvectab mu aber auf ..00-Adresse liegen!
intvectab	db	12 dup(?)	; Interruptvektortabelle, 12 Byte

		if	ubiosend >= intvectab 
		ERROR  "ubiosend >= intvectab"
		endif

stackp:		dw	?		; Zwischenspeicher Stackpointer


ddma:		dw	?		; aktuelle DMA-Adresse
dirbuf:		db	128 dup (?)	; Bereich 128 Byte

ALV0:		db	50 dup (?)	; = DSM/8+1 = 399/8+1 = 50
CSV0:		db	64 dup (?) 	; = CKS, hier max. 256 DIR-Eintrge wg. versch. Formate
ALV1:		db	50 dup (?)
CSV1:		db	64 dup (?)

		IF RAMDSK=1
ALV2:		db	4 dup (?)	; DSM/8+1 = 23/8+1 = 3
		ENDIF

fdcbuffer:	db	400h dup (?) 	; Buffer fr FDC-IO-Operation (1K? 2K?)

ubiosend0:	equ	$

		if	ubiosend0 > RAMEND-30
		ERROR "ubiosend0 > RAMEND"
		endif

biosstack	equ	RAMEND		; Stackbereich	fr BIOS (nur read disk)

; csv unnecessary for non removable media
;csv0   db (DRMC+1)/4 dup (?) = CKS
;alv0   db (DSMC/8)+1 dup (?)

; ENDE BIOS
;-----------------------------------------------------------------------------

ende:		equ	$

		end
